import requests
from requests.adapters import HTTPAdapter
import six
from six.moves.urllib.parse import urlencode
from ckan_cloud_operator.config import manager as config_manager

class StatusCakeError(Exception):
    pass

class StatusCakeAuthError(StatusCakeError):
    pass

class StatusCakeNotLinkedError(StatusCakeError):
    pass

class StatusCakeFieldMissingError(StatusCakeError):
    pass

class StatusCakeFieldError(StatusCakeError):
    pass

class StatusCakeResponseError(StatusCakeError):
    pass


def to_comma_list(value):
    if isinstance(value, (list, tuple, set, frozenset)):
        value = ','.join(value)
    return value


class DeisCkanInstanceUptime(object):
    URL_ALL_TESTS = "https://app.statuscake.com/API/Tests/"
    URL_DETAILS_TEST = "https://app.statuscake.com/API/Tests/Details/?TestID=%s"
    URL_UPDATE_TEST = "https://app.statuscake.com/API/Tests/Update"
    URL_ALL_GROUPS = "https://app.statuscake.com/API/ContactGroups/"

    TESTS_FIELDS = {
        'TestID': (int, None, None),
        'WebsiteName': (six.string_types, None, None),
        'WebsiteURL': (six.string_types, None, None),
        'CheckRate': (int, range(0, 24001), None),
        'TestType': (six.string_types, ("HTTP", "TCP", "PING", "PUSH"), None),
        'ContactGroup': (six.string_types, None, to_comma_list)
    }

    def __init__(self, instance):
        config = config_manager.get(secret_name='uptime-statuscake-api')
        self.statuscake_api_user = config['user']
        self.statuscake_api_key = config['key']
        self.statuscake_group = config['group']
        self.instance_id = instance.id
        self.instance = instance
        self.timeout = 10

        self.session = requests.Session()
        self.session.mount('https://www.statuscake.com', HTTPAdapter(max_retries=5))

    def _request(self, method, url, data=None, auth_headers=True, check_errors=True, **kwargs):
        headers = {}
        if auth_headers:
            headers.update({
                'API': self.statuscake_api_key,
                'Username': self.statuscake_api_user,
            })

        if isinstance(data, dict):
            data = urlencode(data)

        kwargs.setdefault('timeout', self.timeout)
        print_json = kwargs.pop('print_json', False)
        print_raw = kwargs.pop('print_raw', False)
        response = getattr(self.session, method)(url, headers=headers, data=data, **kwargs)
        if print_raw:
            print(response.text)
        if print_json:
            print(response.json())
        if check_errors:
            json_resp = response.json()
            if isinstance(json_resp, dict) and (json_resp.get('Success', True) is False or json_resp.get('Error', None) is not None):
                errno = json_resp.get('ErrNo', -1)
                error_message = json_resp.get('Error')
                if not error_message:
                    error_message = json_resp.get('Message')
                if errno == 0:
                    raise StatusCakeAuthError(error_message or 'Authentication Failed')
                elif errno == 1:
                    raise StatusCakeNotLinkedError(error_message or 'Authentication Failed')
                raise StatusCakeResponseError(error_message or 'API Call Failed')
        return response

    def _check_fields(self, data, check_map):
        for field_name, (field_type, field_values, field_conv) in six.iteritems(check_map):
            if field_name not in data:
                continue
            if field_conv:
                try:
                    data[field_name] = field_conv(data[field_name])
                except TypeError as exc:
                    raise StatusCakeFieldError("Field %s: %s" % (field_name, str(exc)))
            if not isinstance(data[field_name], field_type):
                raise StatusCakeFieldError("Field %s must be of type %s" % (field_name, field_type))
            if field_values is not None and data[field_name] not in field_values:
                raise StatusCakeFieldError("Field %s value %s does not match one of: %s" % (field_name, field_type, field_values))

    def get_contact_groups(self, **kwargs):
        return self._request('get', self.URL_ALL_GROUPS, **kwargs).json()

    def get_all_tests(self, **kwargs):
        return self._request('get', self.URL_ALL_TESTS, **kwargs).json()

    def get_details_test(self, test_id, **kwargs):
        return self._request('get', self.URL_DETAILS_TEST % test_id, **kwargs).json()

    def delete_test(self, test_id, **kwargs):
        return self._request('delete', self.URL_DETAILS_TEST % test_id, **kwargs).json()

    def insert_test(self, data, **kwargs):
        if not isinstance(data, dict):
            raise StatusCakeError("data argument must be a dict")
        if 'WebsiteName' not in data:
            raise StatusCakeFieldMissingError("WebsiteName missing")
        if 'WebsiteURL' not in data:
            raise StatusCakeFieldMissingError("WebsiteURL missing")
        if 'TestType' not in data:
            raise StatusCakeFieldMissingError("TestType missing")
        if 'CheckRate' not in data:
            data['CheckRate'] = 300
        self._check_fields(data, self.TESTS_FIELDS)
        return self._request('put', self.URL_UPDATE_TEST, data=data, **kwargs).json()

    def update_test(self, data, **kwargs):
        if not isinstance(data, dict):
            raise StatusCakeError("data argument must be a dict")
        if 'TestID' not in data:
            raise StatusCakeFieldMissingError("TestID missing")
        # if CheckRate not passed it will be reset to the account plan default (either 30 or 300)
        if 'CheckRate' not in data:
            raise StatusCakeFieldMissingError("CheckRate missing")
        self._check_fields(data, self.TESTS_FIELDS)
        return self._request('put', self.URL_UPDATE_TEST, data=data, **kwargs).json()

    def get_test_id(self, website_name):
        test = list(filter(lambda x: x.get('WebsiteName') == website_name, self.get_all_tests()))
        if not len(test):
            return None
        return test[0]['TestID']

    def update(self, site_url):
        from ckan_cloud_operator.providers.routers import manager as routers_manager
        env_id = routers_manager.get_env_id()
        website_name = f'{env_id}-{self.instance_id}'
        data = {
            "WebsiteName": website_name,
            "WebsiteURL": site_url,
            "CheckRate": 300,
            "TestType": "HTTP",
            "ContactGroup": self.statuscake_group
        }
        test_id = self.get_test_id(website_name)
        if test_id is None:
            try:
                self.insert_test(data)
            except Exception as e:
                print('Failed to create StatusCake test, skipping')
                print(e)
        else:
            data['TestID'] = test_id
            try:
                self.update_test(data)
            except Exception as e:
                print('Failed to update StatusCake test, skipping')
                print(e)


    def delete(self, site_id):
        try:
            test_id = self.get_test_id(self.instance_id)
            self.delete_test(test_id)
        except Exception as e:
            print('Failed to delete StatusCake test, skipping')
            print(e)



import click

from ckan_cloud_operator import logs
from ckan_cloud_operator.providers.db.gcloudsql import manager


@click.group('gcloudsql')
def gcloudsql_group():
    """Manage Google Cloud SQL instances for centralized DB services"""
    pass


@gcloudsql_group.command()
@click.option('--interactive', is_flag=True)
@click.option('--db-prefix')
def initialize(interactive, db_prefix):
    """Enable Google Cloud SQL as the main DB provider"""
    manager.initialize(interactive=interactive, db_prefix=db_prefix)
    logs.exit_great_success()



@gcloudsql_group.command()
@click.argument('OPERATION_ID')
def operation_status(operation_id):
    logs.print_yaml_dump(manager.get_operation_status(operation_id))


@gcloudsql_group.command()
@click.argument('DATABASE')
@click.argument('CONNECTION_STRING', required=False)
@click.option('--db-prefix')
@click.option('--dry-run', is_flag=True)
def create_backup(database, connection_string, db_prefix, dry_run):
    manager.create_backup(database, connection_string, db_prefix=db_prefix, dry_run=dry_run)
    logs.exit_great_success()


@gcloudsql_group.command()
@click.option('--db-prefix')
@click.option('--dry-run', is_flag=True)
def create_all_backups(db_prefix, dry_run):
    manager.create_all_backups(db_prefix, dry_run=dry_run)
    logs.exit_great_success()



from ckan_cloud_operator.routers.routes import deis_instance_subdomain
from ckan_cloud_operator.routers.routes import ckan_instance_subdomain
from ckan_cloud_operator.routers.routes import app_instance_subdomain
from ckan_cloud_operator.routers.routes import datapusher_subdomain
from ckan_cloud_operator.routers.routes import backend_url_subdomain
from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs


def get_module(route):
    route_type = route['spec']['type']
    if route_type == 'deis-instance-subdomain':
        module = deis_instance_subdomain
    elif route_type == 'ckan-instance-subdomain':
        module = ckan_instance_subdomain
    elif route_type == 'app-instance-subdomain':
        module = app_instance_subdomain
    elif route_type == 'datapusher-subdomain':
        module = datapusher_subdomain
    elif route_type == 'backend-url-subdomain':
        module = backend_url_subdomain
    else:
        raise Exception(f'Invalid route type: {route_type}')
    return module


def list(router_labels):
    routes = kubectl.get_items_by_labels('CkanCloudRoute', router_labels, required=False)
    logs.debug_verbose(router_labels=router_labels, routes=routes)
    _routes = []
    if routes:
        for route in routes:
            route = get_module(route).get_route(route)
            logs.debug_verbose(route=route)
            _routes.append(route)
    return _routes


def get_name(route):
    return route['metadata']['name']


def get_backend_url(route):
    return get_module(route).get_backend_url(route)


def get_frontend_hostname(route):
    return get_module(route).get_frontend_hostname(route)


def pre_deployment_hook(route, labels):
    return get_module(route).pre_deployment_hook(route, labels)


def get_domain_parts(route):
    root_domain, sub_domain = get_module(route).get_domain_parts(route)
    return root_domain, sub_domain


def install_crds():
    kubectl.install_crd('ckancloudroutes', 'ckancloudroute', 'CkanCloudRoute')



def get_backend_url(route):
    name, spec = _init_route(route)
    return spec['backend-url']


def get_frontend_hostname(route):
    _, spec = _init_route(route)
    root_domain = spec['root-domain']
    sub_domain = spec['sub-domain']
    return f'{sub_domain}.{root_domain}'


def pre_deployment_hook(route, labels):
    pass


def get_domain_parts(route):
    _, spec = _init_route(route)
    return spec['root-domain'], spec['sub-domain']


def get_default_root_domain():
    from ckan_cloud_operator.providers.routers import manager as routers_manager
    default_root_domain = routers_manager.get_default_root_domain()
    assert default_root_domain, 'missing ckan-infra ROUTERS_DEFAULT_ROOT_DOMAIN'
    return default_root_domain


def get_route(route):
    _init_route(route)
    root_domain = route['spec'].get('root-domain')
    if not root_domain or root_domain == 'default':
        route['spec']['root-domain'] = get_default_root_domain()
    return route


def _init_route(route):
    spec = route['spec']
    name = route['metadata']['name']
    return name, spec



from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.routers.annotations import CkanRoutersAnnotations
from ckan_cloud_operator.infra import CkanInfra
from ckan_cloud_operator.routers.traefik import deployment as traefik_deployment


def create(router):
    router_name = router['metadata']['name']
    router_spec = router['spec']
    cloudflare_spec = router_spec.get('cloudflare', {})
    cloudflare_email = cloudflare_spec.get('email')
    cloudflare_api_key = cloudflare_spec.get('api-key')
    default_root_domain = router_spec.get('default-root-domain')
    dns_provider = router_spec.get('dns-provider')
    default_dns_provider = 'none'
    logs.info(dns_provider=dns_provider, default_dns_provider=default_dns_provider)
    if not dns_provider:
        dns_provider = default_dns_provider
    router_spec['dns-provider'] = dns_provider
    assert all([default_root_domain, dns_provider]), f'invalid traefik router spec: {router_spec}'
    if dns_provider == 'cloudflare':
        assert cloudflare_email and cloudflare_api_key, 'invalid traefik router spec for cloudflare dns provider'
    # cloudflare credentials are stored in a secret, not in the spec
    if 'cloudflare' in router_spec:
        del router_spec['cloudflare']
    kubectl.apply(router)
    annotations = CkanRoutersAnnotations(router_name, router)
    if dns_provider == 'cloudflare':
        annotations.update_flag('letsencryptCloudflareEnabled', lambda: annotations.set_secrets({
            'LETSENCRYPT_CLOUDFLARE_EMAIL': cloudflare_email,
            'LETSENCRYPT_CLOUDFLARE_API_KEY': cloudflare_api_key
        }), force_update=True)
    return router


def delete(router_name):
    print(f'Deleting traefik router {router_name}')
    if all([
        kubectl.call(f'delete --ignore-not-found -l ckan-cloud/router-name={router_name} deployment') == 0,
        kubectl.call(f'delete --ignore-not-found -l ckan-cloud/router-name={router_name} service') == 0,
        kubectl.call(f'delete --ignore-not-found -l ckan-cloud/router-name={router_name} secret') == 0,
        kubectl.call(f'delete --ignore-not-found -l ckan-cloud/router-name={router_name} configmap') == 0,
        kubectl.call(f'delete --ignore-not-found -l ckan-cloud/router-name={router_name} PersistentVolumeClaim') == 0,
        kubectl.call(f'delete --ignore-not-found -l ckan-cloud/router-name={router_name} CkanCloudRoute') == 0,
        kubectl.call(f'delete --ignore-not-found CkanCloudRouter {router_name}') == 0,
    ]):
        print('Removing finalizers')
        success = True
        routes = kubectl.get_items_by_labels('CkanCloudRoute', {'ckan-cloud/router-name': router_name}, required=False)
        if not routes: routes = []
        for route in routes:
            route_name = route['metadata']['name']
            if kubectl.call(
                    f'patch CkanCloudRoute {route_name} -p \'{{"metadata":{{"finalizers":[]}}}}\' --type=merge',
            ) != 0:
                success = False
        if kubectl.get(f'CkanCloudRouter {router_name}', required=False):
            if kubectl.call(
                    f'patch CkanCloudRouter {router_name} -p \'{{"metadata":{{"finalizers":[]}}}}\' --type=merge',
            ) != 0:
                success = False
        assert success
    else:
        raise Exception('Deletion failed')


def get_dns_data(router_name, router, failfast=False):
    return traefik_deployment.get_dns_data(router_name, router, failfast=failfast)


def get(router_name, attr='deployment', router=None, failfast=False):
    deployment_data = lambda: traefik_deployment.get(router_name)
    dns_data = lambda: get_dns_data(router_name, router, failfast=failfast)
    if attr == 'deployment':
        return deployment_data()
    elif attr == 'dns':
        return dns_data()
    else:
        return {'deployment': deployment_data(), 'dns': dns_data()}


def update(router_name, wait_ready, spec, annotations, routes, dry_run=False):
    logs.debug(f'updating traefik router: {router_name}')
    logs.debug_verbose(router_name=router_name, spec=spec, routes=routes)
    return traefik_deployment.update(router_name, wait_ready, spec, annotations, routes, dry_run=dry_run)


def _init_router(router_name):
    router = kubectl.get(f'CkanCloudRouter {router_name}')
    assert router['spec']['type'] == 'traefik'
    annotations = CkanRoutersAnnotations(router_name, router)
    ckan_infra = CkanInfra()
    return router, annotations, ckan_infra



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False, suffix=None): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment, suffix=suffix)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False, interactive=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file, interactive)

################################
# custom provider code starts here
#

import os
import binascii
import yaml
import json

from ckan_cloud_operator import kubectl
from ckan_cloud_operator.drivers.kubectl import rbac
from ckan_cloud_operator import logs
from ckan_cloud_operator.providers.cluster.aws import manager as aws_manager


def initialize(interactive=False, dry_run=False):
    aws_creds = aws_manager.get_aws_credentials()
    _config_interactive_set({
        'file.system.id': None,
    }, interactive=interactive, namespace='default')
    for key, value in [
        ('aws.region', aws_creds['region']),
        ('provisioner.name', 'example.com/aws-efs'),
        # ('dns.name', ''),
    ]:
        _config_set(key=key, value=value, namespace='default')
    _apply_rbac()
    _apply_deployment()


def _apply_rbac():
    labels = _get_resource_labels()
    rbac.update_service_account('efs-provisioner', labels, 'default')
    rbac.update_cluster_role('efs-provisioner-runner', [
        dict(zip(['apiGroups', 'resources', 'verbs'], rule))
        for rule in (
            ([''], ['persistentvolumes'], ['get', 'list', 'watch', 'create', 'delete']),
            ([''], ['persistentvolumeclaims'], ['get', 'list', 'watch', 'update']),
            (['storage.k8s.io'], ['storageclasses'], ['get', 'list', 'watch']),
            ([''], ['events'], ['create', 'update', 'patch']),
        )
    ], labels)
    rbac.update_cluster_role_binding('run-efs-provisioner',  dict(
            kind='ServiceAccount', name='efs-provisioner', namespace='default',
    ), 'efs-provisioner-runner', labels)
    rbac.update_role('leader-locking-efs-provisioner', labels, [dict(
            apiGroups=[''],
            resources=['endpoints'],
            verbs=['get', 'list', 'watch', 'create', 'update', 'patch']
    )], 'default')
    rbac.update_role_binding('leader-locking-efs-provisioner', 'leader-locking-efs-provisioner', 'default', 'efs-provisioner', labels)

def _apply_deployment():
    labels = _get_resource_labels(for_deployment=True)
    labels = dict(labels, app='efs-provisioner')
    configmap_name = _get_resource_name()
    fs_id = _config_get('file.system.id', namespace='default')
    aws_region = _config_get('aws.region', namespace='default')
    kubectl.apply(kubectl.get_deployment(
        'efs-provisioner',
        labels,
        dict(
            replicas=1,
            strategy=dict(type='Recreate'),
            selector=dict(matchLabels=labels),
            template=dict(
                metadata=dict(labels=labels),
                spec=dict(
                    containers=[dict(
                        name='efs-provisioner',
                        image='quay.io/external_storage/efs-provisioner:latest',
                        env=[
                            dict(
                                name=name,
                                valueFrom=dict(
                                    configMapKeyRef=dict(
                                        name=configmap_name,
                                        key=key,
                                        optional=optional
                                    )
                                )
                            )
                            for name, key, optional in [
                                ('FILE_SYSTEM_ID', 'file.system.id', False),
                                ('AWS_REGION', 'aws.region', False),
                                ('DNS_NAME', 'dns.name', True),
                                ('PROVISIONER_NAME', 'provisioner.name', False),
                            ]
                        ],
                        volumeMounts=[dict(
                            name='pv-volume',
                            mountPath='/persistentvolumes'
                        )]
                    )],
                    volumes=[dict(
                        name='pv-volume',
                        nfs=dict(
                            server=f'{fs_id}.efs.{aws_region}.amazonaws.com',
                            path='/'
                        )
                    )]
                )
            )
        ),
        namespace='default'
    ))



from setuptools import setup, find_packages
from os import path, environ
from time import time

here = path.abspath(path.dirname(__file__))

if path.exists("VERSION.txt"):
    # this file can be written by CI tools (e.g. Travis)
    with open("VERSION.txt") as version_file:
        version = version_file.read().strip().strip("v")
else:
    version = str(time())

req_file_name = 'requirements.min.in' if environ.get('INSTALL_MINIFIED') else 'requirements.in'

with open(req_file_name) as requirements_file:
    install_requires = requirements_file.read().strip().split('\n')

setup(
    name='cco',
    version=version,
    description='''CKAN Cloud Kubernetes operator''',
    url='https://github.com/datopian/ckan-cloud-operator',
    author='''Viderum''',
    license='MIT',
    packages=find_packages(exclude=['examples', 'tests', '.tox']),
    install_requires=install_requires,
    entry_points={
      'console_scripts': [
        'ckan-cloud-operator = ckan_cloud_operator.cli:main',
        'cco = ckan_cloud_operator.cli:main',
      ]
    },
)



import yaml
import datetime


### disable yaml load warnings
# see https://github.com/yaml/pyyaml/wiki/PyYAML-yaml.load(input)-Deprecation

if hasattr(yaml, 'warnings'):
    yaml.warnings({'YAMLLoadWarning': False})


### global fix for yaml handling of datetime formats with compatibility for kubectl


datetime_format = '%Y-%m-%dT%H:%M:%SZ'


def datetime_representer(dumper, data):
    return dumper.represent_data(data.strftime(datetime_format))


def datetime_constructor(loader, node):
    value = loader.construct_scalar(node)
    return datetime.datetime.strptime(value, datetime_format)


yaml.add_representer(datetime.datetime, datetime_representer)
yaml.add_constructor(u'!datetime', datetime_constructor)


def yaml_dump(data, file, *args, **kwargs):
    return yaml.dump(data, file, *args, Dumper=YamlSafeDumper, default_flow_style=False, **kwargs)


class YamlSafeDumper(yaml.SafeDumper):

    def ignore_aliases(self, data):
        return True

    def represent_undefined(self, data):
        return None



from ckan_cloud_operator import kubectl
from ckan_cloud_operator.infra import CkanInfra


from ckan_cloud_operator.routers.routes.backend_url_subdomain import (
    get_frontend_hostname,
    get_domain_parts,
    get_default_root_domain,
    get_route,
    _init_route
)


def get_backend_url(route):
    name, spec = _init_route(route)
    deis_instance_id = spec['deis-instance-id']
    target_port = _get_instance_target_port(deis_instance_id)
    return f'http://{name}.{deis_instance_id}:{target_port}'


def pre_deployment_hook(route, labels):
    name, spec = _init_route(route)
    deis_instance_id = spec['deis-instance-id']
    if kubectl.get(f'ns {deis_instance_id}', required=False):
        target_port = _get_instance_target_port(deis_instance_id)
        print(f'updating route name {name} for deis instance {deis_instance_id} (port={target_port})')
        route_service = kubectl.get_resource('v1', 'Service', name, labels, namespace=deis_instance_id)
        route_service['spec'] = {
            'ports': [
                {'name': str(target_port), 'port': target_port}
            ],
            'selector': {
                'app': 'ckan'
            }
        }
        kubectl.apply(route_service)


def _get_instance_target_port(instance_id):
    target_port = 5000
    instance = kubectl.get(f'ckancloudckaninstance {instance_id}', required=False)
    if instance:
        _target_port = instance.get('spec', {}).get('routes', {}).get('target-port')
        if _target_port:
            target_port = _target_port
    return target_port



import psycopg2
import contextlib
import traceback

from ckan_cloud_operator import logs


@contextlib.contextmanager
def connect(*args, **kwargs):
    with psycopg2.connect(*args, **kwargs) as conn:
        yield conn


def create_base_db(admin_conn, db_name, db_password, grant_to_user=None):
    db_info = get_db_role_info(admin_conn, db_name)
    errors = []
    if db_info.get('role'):
        logs.info(f'Role already exists: {db_name}')
        errors.append('role-exists')
    else:
        create_role_if_not_exists(admin_conn, db_name, db_password)
    if db_info.get('db'):
        logs.info(f'DB already exists: {db_name}')
        errors.append('db-exists')
    else:
        logs.info(f'Creating DB: {db_name}')
        _set_session_autocommit(admin_conn)
        with admin_conn.cursor() as cur:
            cur.execute(f'CREATE DATABASE "{db_name}";')
        _unset_session_autocommit(admin_conn)
    if grant_to_user:
        _set_session_autocommit(admin_conn)
        with admin_conn.cursor() as cur:
            cur.execute(f'GRANT "{db_name}" to "{grant_to_user}";')
        _unset_session_autocommit(admin_conn)

    return errors


def create_role_if_not_exists(admin_conn, role_name, role_password):
    roles = list(list_roles(admin_conn, role_name=role_name))
    if len(roles) == 0:
        logs.info(f'Creating role: {role_name}')
        _set_session_autocommit(admin_conn)
        with admin_conn.cursor() as cur:
            cur.execute(f'CREATE ROLE "{role_name}" WITH LOGIN PASSWORD %s NOSUPERUSER NOCREATEDB NOCREATEROLE;',
                        (role_password,))
            cur.execute(f'GRANT "{role_name}" TO postgres;')
        _unset_session_autocommit(admin_conn)


def delete_role(admin_conn, role_name):
    roles = list(list_roles(admin_conn, role_name=role_name))
    if len(roles) > 0:
        logs.info(f'Deleting role: {role_name}')
        _set_session_autocommit(admin_conn)
        with admin_conn.cursor() as cur:
            try:
                cur.execute(f'DROP ROLE "{role_name}"')
            except psycopg2.ProgrammingError:
                traceback.print_exc()
        _unset_session_autocommit(admin_conn)


def delete_base_db(admin_conn, db_name):
    errors = []
    db_info = get_db_role_info(admin_conn, db_name)
    logs.info(f'Revoking connect and terminating all connections db: {db_name}')
    _set_session_autocommit(admin_conn)
    with admin_conn.cursor() as cur:
        try: cur.execute(f'REVOKE CONNECT ON DATABASE "{db_name}" FROM public;')
        except psycopg2.ProgrammingError:
            traceback.print_exc()
    _unset_session_autocommit(admin_conn)
    with admin_conn.cursor() as cur:
        try: cur.execute('SELECT pg_terminate_backend(pg_stat_activity.pid) '
                         'FROM pg_stat_activity '
                         'WHERE pg_stat_activity.datname = %s;', (db_name,))
        except psycopg2.ProgrammingError:
            traceback.print_exc()
    if db_info.get('db'):
        logs.info(f'Deleting db: {db_name}')
        _set_session_autocommit(admin_conn)
        with admin_conn.cursor() as cur:
            try: cur.execute(f'DROP DATABASE "{db_name}"')
            except psycopg2.ProgrammingError:
                traceback.print_exc()
        _unset_session_autocommit(admin_conn)
    else:
        logs.info(f'DB does not exist: {db_name}')
        errors.append('db-does-not-exist')
    if db_info.get('role'):
        logs.info(f'Deleting role: {db_name}')
        _set_session_autocommit(admin_conn)
        with admin_conn.cursor() as cur:
            try:
                cur.execute(f'DROP ROLE "{db_name}"')
            except psycopg2.ProgrammingError:
                traceback.print_exc()
        _unset_session_autocommit(admin_conn)
    else:
        logs.info(f'Role does not exist: {db_name}')
        errors.append('role-does-not-exist')
    return errors


def get_db_role_info(admin_conn, db_name):
    res = {'role': None, 'db': None}
    cur = admin_conn.cursor()
    fields = 'rolname | rolsuper | rolinherit | rolcreaterole | rolcreatedb | rolcanlogin | rolreplication | rolconnlimit | rolpassword | rolvaliduntil | rolbypassrls | rolconfig |  oid'.split(' | ')
    fields_select = ', '.join(fields)
    cur.execute(f'select {fields_select} from pg_roles where rolname=%s', (db_name,))
    row = cur.fetchone()
    res['role'] = dict(zip(fields, row)) if row else None
    fields = 'datname | datdba | encoding | datcollate | datctype | datistemplate | datallowconn | datconnlimit | datlastsysoid | datfrozenxid | datminmxid | dattablespace | datacl'.split(' | ')
    fields_select = ', '.join(fields)
    cur.execute(f'select {fields_select} from pg_database where datname=%s', (db_name,))
    row = cur.fetchone()
    res['db'] = dict(zip(fields, row)) if row else None
    cur.close()
    return res


def list_db_names(admin_conn, full=False, validate=False):
    if validate: full = True
    failures = []
    cur = admin_conn.cursor()
    cur.execute('select datname from pg_database')
    for row in cur:
        db_name = row[0]
        if full:
            data = get_db_role_info(admin_conn, db_name)
            if validate and not (data.get('db') and data.get('role')): failures.append(db_name)
            yield data
        else:
            yield db_name
    if validate and len(failures) > 0: raise Exception(f'Failed to get role for following dbs: {failures}')


def list_roles(admin_conn, full=False, validate=False, role_name=None):
    if validate: full=True
    failures = []
    cur = admin_conn.cursor()
    if role_name is not None:
        where = ' where rolname=%s'
        args = (role_name,)
    else:
        where = ''
        args = ()
    cur.execute(f'select rolname from pg_roles{where}', args)
    for row in cur:
        role_name = row[0]
        if full:
            data = get_db_role_info(admin_conn, role_name)
            if validate and not (data.get('db') and data.get('role')): failures.append(role_name)
            yield data
        else:
            yield role_name
    if validate and len(failures) > 0: raise Exception(f'Failed to get db for following roles: {failures}')


def initialize_extensions(admin_db_conn, extension_names):
    with admin_db_conn.cursor() as cur:
        cur.execute(' '.join([
            f'CREATE EXTENSION IF NOT EXISTS {extension_name};'
            for extension_name in extension_names
        ]))


def _set_session_autocommit(conn):
    conn.commit()
    conn.set_session(autocommit=True)


def _unset_session_autocommit(conn):
    conn.commit()
    conn.set_session(autocommit=False)



from ckan_cloud_operator import kubectl


def get(what, *args, required=True, namespace=None, get_cmd=None, **kwargs):
    return kubectl.get(what, *args, required=required, namespace=namespace, get_cmd=get_cmd, **kwargs)



import subprocess
import yaml
import sys
import datetime
from urllib.parse import urlparse
import os

from ckan_cloud_operator import kubectl


class CkanInfra(object):

    @classmethod
    def add_cli_commands(cls, click, command_group, great_success):

        @command_group.command('clone')
        def __ckan_infra_clone():
            """Clone the infrastructure secret from an existing secret piped on stdin

            Example: KUBECONFIG=/other/.kube-config kubectl -n ckan-cloud get secret ckan-infra -o yaml | ckan-cloud-operator ckan-infra clone
            """
            cls.clone(yaml.load(sys.stdin.read()))
            great_success()

        @command_group.group('set')
        def __ckan_infra_set():
            """Set or overwrite infrastructure secrets"""
            pass

        @__ckan_infra_set.command('key-value')
        @click.argument('KEY')
        @click.argument('VALUE')
        def ckan_infra_set_key_value(key, value):
            """Sets an arbitrary key/value pair on the ckan-infra secret. Use --force to overwrite existing values"""
            cls.set('key-value', key, value)
            great_success()


        @__ckan_infra_set.command('gcloud')
        @click.argument('GCLOUD_SERVICE_ACCOUNT_JSON_FILE')
        @click.argument('GCLOUD_SERVICE_ACCOUNT_EMAIL')
        @click.argument('GCLOUD_AUTH_PROJECT')
        def ckan_infra_set_gcloud(gcloud_service_account_json_file, gcloud_service_account_email, gcloud_auth_project):
            """Sets the Google cloud authentication details, should run locally or mount the json file into the container"""
            cls.set('gcloud', gcloud_service_account_json_file, gcloud_service_account_email, gcloud_auth_project)
            great_success()

        @__ckan_infra_set.command('docker-registry')
        @click.argument('DOCKER_REGISTRY_SERVER')
        @click.argument('DOCKER_REGISTRY_USERNAME')
        @click.argument('DOCKER_REGISTRY_PASSWORD')
        @click.argument('DOCKER_REGISTRY_EMAIL')
        def ckan_infra_set_docker_registry(docker_registry_server, docker_registry_username, docker_registry_password,
                                           docker_registry_email):
            """Sets the Docker registry details for getting private images for CKAN pods in the cluster"""
            cls.set('docker-registry',
                    docker_registry_server, docker_registry_username, docker_registry_password, docker_registry_email)
            great_success()

        @command_group.command('get')
        @click.argument('CKAN_INFRA_KEY', required=False)
        def ckan_infra_get(ckan_infra_key):
            """Get the ckan-infra secrets"""
            if ckan_infra_key:
                print(getattr(cls(), ckan_infra_key))
            else:
                print(yaml.dump(cls.get(), default_flow_style=False))

        @command_group.command('admin-db-connection-string')
        def ckan_infra_admin_db_connection_string():
            """Get a DB connection string for administration

            Example: psql -d $(ckan-cloud-operator admin-db-connection-string)
            """
            infra = cls()
            postgres_user = infra.POSTGRES_USER
            postgres_password = infra.POSTGRES_PASSWORD
            if os.environ.get('CKAN_CLOUD_OPERATOR_USE_PROXY') in ['yes', '1', 'true']:
                postgres_host = '127.0.0.1'
            else:
                postgres_host = infra.POSTGRES_HOST
            postgres_port = '5432'
            print(f'postgresql://{postgres_user}:{postgres_password}@{postgres_host}:{postgres_port}')

        @command_group.command('cloudsql-proxy')
        def ckan_infra_cloudsql_proxy():
            """Starts a local proxy to the cloud SQL instance"""
            print("Keep this running in the background")
            print("Set the following environment variable to cause ckan-cloud-operator to connect via the proxy")
            print("export CKAN_CLOUD_OPERATOR_USE_PROXY=yes")
            subprocess.check_call(f'kubectl -n ckan-cloud port-forward deployment/ckan-cloud-db-proxy-pgbouncer 5432', shell=True)

        @command_group.command('deploy-solr-proxy')
        def deploy_ckan_infra_solr_proxy():
            """Deploys a proxy inside the cluster which allows to access the centralized solr without authentication"""
            labels = {'app': 'ckan-cloud-solrcloud-proxy'}
            infra = cls()
            solr_url = urlparse(infra.SOLR_HTTP_ENDPOINT)
            scheme = solr_url.scheme
            hostname = solr_url.hostname
            port = solr_url.port
            if not port:
                port = '443' if scheme == 'https' else '8983'
            kubectl.update_secret('solrcloud-proxy', {
                'SOLR_URL': f'{scheme}://{hostname}:{port}',
                'SOLR_USER': infra.SOLR_USER,
                'SOLR_PASSWORD': infra.SOLR_PASSWORD
            })
            kubectl.apply(kubectl.get_deployment('solrcloud-proxy', labels, {
                'replicas': 1,
                'revisionHistoryLimit': 10,
                'strategy': {'type': 'RollingUpdate', },
                'template': {
                    'metadata': {
                        'labels': labels,
                        'annotations': {
                            'ckan-cloud/operator-timestamp': str(datetime.datetime.now())
                        }
                    },
                    'spec': {
                        'containers': [
                            {
                                'name': 'solrcloud-proxy',
                                'image': 'viderum/ckan-cloud-operator-solrcloud-proxy',
                                'envFrom': [{'secretRef': {'name': 'solrcloud-proxy'}}],
                                'ports': [{'containerPort': 8983}],
                            }
                        ]
                    }
                }
            }))
            service = kubectl.get_resource('v1', 'Service', 'solrcloud-proxy', labels)
            service['spec'] = {
                'ports': [
                    {'name': '8983', 'port': 8983}
                ],
                'selector': labels
            }
            kubectl.apply(service)

    def __init__(self, required=True):
        if required:
            values = kubectl.decode_secret(kubectl.get('secret ckan-infra', required=required), required=required)
        else:
            values = {}

        # Database

        self.POSTGRES_HOST = values.get('POSTGRES_HOST')
        self.POSTGRES_USER = values.get('POSTGRES_USER')
        self.POSTGRES_PASSWORD = values.get('POSTGRES_PASSWORD')
        self.GCLOUD_SQL_INSTANCE_NAME = values.get('GCLOUD_SQL_INSTANCE_NAME')
        self.GCLOUD_SQL_PROJECT = values.get('GCLOUD_SQL_PROJECT')

        # Solr

        self.SOLR_HTTP_ENDPOINT = values.get('SOLR_HTTP_ENDPOINT')
        self.SOLR_HTTP_ENDPOINT_SIMPLE = values.get('SOLR_HTTP_ENDPOINT_SIMPLE')
        self.SOLR_USER = values.get('SOLR_USER')
        self.SOLR_PASSWORD = values.get('SOLR_PASSWORD')
        self.SOLR_NUM_SHARDS = values.get('SOLR_NUM_SHARDS')
        self.SOLR_REPLICATION_FACTOR = values.get('SOLR_REPLICATION_FACTOR')

        # Private Docker Registry

        self.DOCKER_REGISTRY_SERVER = values.get('DOCKER_REGISTRY_SERVER')
        self.DOCKER_REGISTRY_USERNAME = values.get('DOCKER_REGISTRY_USERNAME')
        self.DOCKER_REGISTRY_PASSWORD = values.get('DOCKER_REGISTRY_PASSWORD')
        self.DOCKER_REGISTRY_EMAIL = values.get('DOCKER_REGISTRY_EMAIL')

        # GitLab

        self.GITLAB_TOKEN_USER = values.get('GITLAB_TOKEN_USER')
        self.GITLAB_TOKEN_PASSWORD = values.get('GITLAB_TOKEN_PASSWORD')

        # Migration from old cluster

        self.DEIS_KUBECONFIG = values.get('DEIS_KUBECONFIG')
        self.GCLOUD_SQL_DEIS_IMPORT_BUCKET = values.get('GCLOUD_SQL_DEIS_IMPORT_BUCKET')

        # Gcloud credentials / general details

        self.GCLOUD_SERVICE_ACCOUNT_JSON = values.get('GCLOUD_SERVICE_ACCOUNT_JSON')
        self.GCLOUD_SERVICE_ACCOUNT_EMAIL = values.get('GCLOUD_SERVICE_ACCOUNT_EMAIL')
        self.GCLOUD_AUTH_PROJECT = values.get('GCLOUD_AUTH_PROJECT')
        self.GCLOUD_COMPUTE_ZONE = values.get('GCLOUD_COMPUTE_ZONE')
        self.GCLOUD_CLUSTER_NAME = values.get('GCLOUD_CLUSTER_NAME')

        # Gcloud Storage

        self.GCLOUD_STORAGE_BUCKET = values.get('GCLOUD_STORAGE_BUCKET')
        self.GCLOUD_STORAGE_ACCESS_KEY_ID = values.get('GCLOUD_STORAGE_ACCESS_KEY_ID')
        self.GCLOUD_STORAGE_SECRET_ACCESS_KEY = values.get('GCLOUD_STORAGE_SECRET_ACCESS_KEY')
        self.GCLOUD_STORAGE_HOST_NAME = values.get('GCLOUD_STORAGE_HOST_NAME')
        self.GCLOUD_STORAGE_REGION_NAME = values.get('GCLOUD_STORAGE_REGION_NAME')
        self.GCLOUD_STORAGE_SIGNATURE_VERSION = values.get('GCLOUD_STORAGE_SIGNATURE_VERSION')

        # Cluster Storage

        self.MULTI_USER_STORAGE_CLASS_NAME = values.get('MULTI_USER_STORAGE_CLASS_NAME', 'cca-ckan')

        # Routers / Load Balancing

        self.ROUTERS_ENV_ID = values.get('ROUTERS_ENV_ID')
        self.ROUTERS_DEFAULT_ROOT_DOMAIN = values.get('ROUTERS_DEFAULT_ROOT_DOMAIN')
        self.ROUTERS_DEFAULT_CLOUDFLARE_EMAIL = values.get('ROUTERS_DEFAULT_CLOUDFLARE_EMAIL')
        self.ROUTERS_DEFAULT_CLOUDFLARE_AUTH_KEY = values.get('ROUTERS_DEFAULT_CLOUDFLARE_AUTH_KEY')

        # Monitoring
        self.CKAN_STATUSCAKE_API_KEY = values.get('CKAN_STATUSCAKE_API_KEY')
        self.CKAN_STATUSCAKE_API_USER = values.get('CKAN_STATUSCAKE_API_USER')
        self.CKAN_STATUSCAKE_GROUP = values.get('CKAN_STATUSCAKE_GROUP')

    @classmethod
    def set(cls, set_type, *args):
        print(f'Setting {set_type} infra secrets')
        if set_type == 'key-value':
            key, value = args
            kubectl.update_secret('ckan-infra', {key: value})
        elif set_type == 'gcloud':
            service_account_json, service_account_email, auth_project = args
            with open(service_account_json, 'rb') as f:
                kubectl.update_secret('ckan-infra',
                                      {'GCLOUD_SERVICE_ACCOUNT_JSON': f.read().decode(),
                                       'GCLOUD_SERVICE_ACCOUNT_EMAIL': service_account_email,
                                       'GCLOUD_AUTH_PROJECT': auth_project})
        elif set_type == 'docker-registry':
            server, username, password, email = args
            kubectl.update_secret('ckan-infra',
                                  {'DOCKER_REGISTRY_SERVER': server,
                                   'DOCKER_REGISTRY_USERNAME': username,
                                   'DOCKER_REGISTRY_PASSWORD': password,
                                   'DOCKER_REGISTRY_EMAIL': email})

        else:
            raise NotImplementedError(f'Invalid infra set spec: {set_type}={args}')

    @classmethod
    def clone(cls, other_secret):
        secret = {
            'apiVersion': 'v1',
            'kind': 'Secret',
            'metadata': {
                'name': 'ckan-infra',
                'namespace': 'ckan-cloud'
            },
            'type': 'Opaque',
            'data': other_secret['data']
        }
        subprocess.run('kubectl create -f -', input=yaml.dump(secret).encode(),
                       shell=True, check=True)

    @classmethod
    def get(cls):
        secret = kubectl.get('secret ckan-infra', required=False)
        if secret:
            return kubectl.decode_secret(secret)
        else:
            return {}



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#

import datetime
import os
import yaml
import subprocess
import json

from ckan_cloud_operator import logs
from ckan_cloud_operator.providers.cluster import manager as cluster_manager


def initialize(db_prefix=None, interactive=False):
    _set_provider()
    default_values = {
        'azuresql-instance-name': '',
        'azuresql-host': '',
        'admin-user': '',
        'admin-password': ''
    }
    if interactive:
        print("\n"
              "Starting interactive initialization of the Azure SQL db provider\n"
              "Please prepare the following values:\n"
              "\n"
              " - admin db host and credentials\n"
              " - SQL instance name\n"
              "\n")
        _config_interactive_set(default_values, **_get_config_credentials_kwargs(db_prefix))
    config = _config_get(**_get_config_credentials_kwargs(db_prefix))
    for key, default_value in default_values.items():
        if key not in config and default_value:
            _config_set(key, default_value, **_get_config_credentials_kwargs(db_prefix))
        elif not config.get(key):
            raise Exception(f'missing key: {key}')
    from ckan_cloud_operator.providers.db.proxy.azuresql import manager as proxy_azuresql_manager
    proxy_azuresql_manager.initialize(db_prefix=db_prefix)


def _get_config_credentials_kwargs(db_prefix):
    return {
        'is_secret': True,
        'suffix': f'{db_prefix}-credentials' if db_prefix else 'credentials'
    }


def is_private_ip():
    return False


def get_postgres_admin_credentials(db_prefix=None):
    credentials = _credentials_get(db_prefix)
    return credentials['admin-user'], credentials['admin-password'], credentials.get('admin-db-name', credentials['admin-user'])


def _credentials_get(db_prefix, key=None, default=None, required=False):
    return _config_get(key=key, default=default, required=required, **_get_config_credentials_kwargs(db_prefix))


def get_postgres_external_host_port(db_prefix=None):
    config = _config_get(**_get_config_credentials_kwargs(db_prefix))
    return config['azuresql-host'], 5432


def get_postgres_internal_host_port(db_prefix=None):
    return get_postgres_external_host_port(db_prefix)



import yaml
import time
import datetime
import binascii
import os
import traceback
import subprocess
import sys

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.labels import manager as labels_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager
from ckan_cloud_operator.providers.storage.constants import CONFIG_NAME
from ckan_cloud_operator.providers.storage.manager import get_provider_id as get_storage_provider_id
from ckan_cloud_operator.routers import manager as routers_manager

from ..constants import INSTANCE_NAME_CRD_SINGULAR
from ..constants import INSTANCE_CRD_SINGULAR
from ..deployment import manager as deployment_manager


def create(instance_type, instance_id=None, instance_name=None, values=None, values_filename=None, exists_ok=False,
           dry_run=False, update_=False, wait_ready=False, skip_deployment=False, skip_route=False, force=False):
    if not instance_id:
        if instance_name:
            instance_id = '{}-{}'.format(instance_name, _generate_password(6))
        else:
            instance_id = _generate_password(12)
    if values_filename:
        assert values is None
        if values_filename != '-':
            with open(values_filename) as f:
                values = yaml.load(f.read())
        else:
            values = yaml.load(sys.stdin.read())
    instance = crds_manager.get(INSTANCE_CRD_SINGULAR, name=instance_id, required=False)
    if not exists_ok and instance:
        raise Exception('instance already exists')
    values_id = values.get('id')
    if values_id and values_id != instance_id:
        logs.warning(f'changing instance id in spec from {values_id} to the instance id {instance_id}')
    values['id'] = instance_id

    use_cloud_storage = values.get('useCloudStorage') and config_manager.get('use-cloud-native-storage', secret_name=CONFIG_NAME)
    values['useCloudStorage'] = use_cloud_storage

    if instance:
        logs.info('Updating instance', instance_id=instance_id)
        instance['spec'] = values
        kubectl.apply(instance, dry_run=dry_run)
    else:
        logs.info('Creating instance', instance_id=instance_id)
        kubectl.apply(crds_manager.get_resource(
            INSTANCE_CRD_SINGULAR, instance_id,
            extra_label_suffixes={'instance-type': instance_type},
            spec=values
        ), dry_run=dry_run)

    if instance_name:
        set_name(instance_id, instance_name, dry_run=dry_run)

    if use_cloud_storage:
        set_storage(instance_id, instance_name, dry_run=dry_run)

    if update_:
        update(instance_id, wait_ready=wait_ready, skip_deployment=skip_deployment, skip_route=skip_route, force=force,
               dry_run=dry_run)

    return instance_id


def update(instance_id_or_name, override_spec=None, persist_overrides=False, wait_ready=False, skip_deployment=False,
           skip_route=False, force=False, dry_run=False):
    instance_id, instance_type, instance = _get_instance_id_and_type(instance_id_or_name, required=not dry_run)
    if dry_run:
        logs.info('update instance', instance_id=instance_id, instance_id_or_name=instance_id_or_name,
                  override_spec=override_spec, persist_overrides=persist_overrides, wait_ready=wait_ready,
                  skip_deployment=skip_deployment, skip_route=skip_route, force=force, dry_run=dry_run)
    else:
        pre_update_hook_data = deployment_manager.pre_update_hook(instance_id, instance_type, instance, override_spec,
                                                                  skip_route)

        bucket_credentials = instance['spec'].get('ckanStorageBucket', {}).get(get_storage_provider_id())
        use_cloud_storage = bucket_credentials and config_manager.get('use-cloud-native-storage', secret_name=CONFIG_NAME)

        if use_cloud_storage:
            cluster_provider_id = cluster_manager.get_provider_id()

            if bucket_credentials:
                literal = []
                config_manager.set(
                    values=bucket_credentials,
                    secret_name='bucket-credentials',
                    namespace=instance_id
                )

        if persist_overrides:
            logs.info('Persisting overrides')
            kubectl.apply(instance)
        if not skip_deployment:
            deployment_manager.update(instance_id, instance_type, instance, force=force)
            if wait_ready:
                wait_instance_ready(instance_id_or_name)
        if not skip_route and pre_update_hook_data.get('sub-domain'):
            root_domain = pre_update_hook_data.get('root-domain')
            sub_domain = pre_update_hook_data['sub-domain']
            assert root_domain == routers_manager.get_default_root_domain(), 'invalid domain, must use default root domain'
            logs.info(f'adding instance default route to {sub_domain}.{root_domain}')
            routers_manager.create_subdomain_route('instances-default', {
                'target-type': 'ckan-instance',
                'ckan-instance-id': instance_id,
                'root-domain': root_domain,
                'sub-domain': sub_domain
            })
            logs.info(f'updating routers_manager wait_ready: {wait_ready}')
            routers_manager.update('instances-default', wait_ready)
        else:
            logs.info('skipping route creation', skip_route=skip_route, sub_domain=pre_update_hook_data.get('sub-domain'))
        logs.info('creating ckan admin')
        # Need to set in values.yaml
        if pre_update_hook_data.get('create-sysadmin'):
            ckan_admin_email = pre_update_hook_data.get('ckan-admin-email')
            ckan_admin_password = pre_update_hook_data.get('ckan-admin-password')
            ckan_admin_name = pre_update_hook_data.get('ckan-admin-name', 'admin')
            res = create_ckan_admin_user(instance_id, ckan_admin_name, ckan_admin_email, ckan_admin_password)
            logs.info(**res)
        logs.info('Instance is ready', instance_id=instance_id, instance_name=(instance_id_or_name if instance_id_or_name != instance_id else None))


def delete(instance_id):
    try:
        instance_id, instance_type, instance = _get_instance_id_and_type(instance_id=instance_id)
    except Exception:
        logs.error(traceback.format_exc())
        instance, instance_type = None, None
    try:
        deployment_manager.delete(instance_id, instance_type, instance)
    except Exception as e:
        logs.error('error during deployment delete', error=str(e))
    try:
        crds_manager.delete(INSTANCE_CRD_SINGULAR, instance_id)
    except Exception as e:
        logs.error('error during crd delete', error=str(e))


def delete_instances(instance_ids_or_names=None, dry_run=True, instance_ids=None):
    if instance_ids:
        assert not instance_ids_or_names
        use_instance_ids = True
    else:
        assert not instance_ids
        use_instance_ids = False
    logs.info(use_instance_ids=use_instance_ids, instance_ids=instance_ids, instance_ids_or_names=instance_ids_or_names)
    for instance_id_or_name in (instance_ids if use_instance_ids else instance_ids_or_names):
        logs.info(instance_id_or_name=instance_id_or_name)
        if use_instance_ids:
            instance_id = instance_id_or_name
        else:
            instance_id, _, _ = _get_instance_id_and_type(instance_id_or_name)
        instance_name = instance_id_or_name if instance_id != instance_id_or_name else None
        yield {
            'id': instance_id,
            **({'name': instance_name} if instance_name else {})
        }
        errors = []
        if instance_id:
            if not dry_run:
                try:
                    delete(instance_id=instance_id)
                except Exception as e:
                    errors.append(str(e))
                    logs.error('exception raised during instance deletion, '
                               'this does not always indicate deletion was not successful',
                               exception=str(e))
        else:
            errors.append(f'Failed to get instance_id for instance_id_or_name {instance_id_or_name}')
        if instance_name:
            try:
                delete_name(instance_name)
            except Exception as e:
                errors.append(str(e))
                logs.error('exception raised during instance name deletion.', exception=str(e))
        yield {'errors': errors}


def wait_instance_ready(instance_id_or_name):
    logs.info(f'Waiting for instance ready status ({instance_id_or_name})')
    time.sleep(3)
    while True:
        data = get(instance_id_or_name)
        if data.get('ready'):
            break
        else:
            logs.print_yaml_dump(
                {
                    k: v for k, v in data.items()
                    if (k not in ['ready'] and type(v) == dict and not v.get('ready')) or k == 'namespace'
                }
            )
            time.sleep(2)


def edit(instance_id_or_name):
    instance_id, instance_type, instance = _get_instance_id_and_type(instance_id_or_name)
    crds_manager.edit(INSTANCE_CRD_SINGULAR, name=instance_id)


def get(instance_id_or_name, attr=None, exclude_attr=None, with_spec=False):
    """Get detailed information about the instance and related components"""
    instance_id, instance_type, instance = _get_instance_id_and_type(instance_id_or_name)
    if not exclude_attr:
        exclude_attr = []
    if not with_spec:
        exclude_attr.append('spec')
    gets = {
        'deployment': lambda: deployment_manager.get(instance_id, instance_type, instance),
        'spec': lambda: instance['spec']
    }
    if exclude_attr:
        gets = {k: v for k, v in gets.items() if k not in exclude_attr}
    if attr:
        return gets[attr]()
    else:
        ret = {'ready': True}
        for k, v in gets.items():
            ret[k] = v()
            if type(ret[k]) == dict and not ret[k].get('ready'):
                ret['ready'] = False
        ret['id'] = instance_id
        return ret


def get_backend_url(instance_id_or_name=None, instance_id=None):
    return deployment_manager.get_backend_url(*_get_instance_id_and_type(instance_id_or_name, instance_id=instance_id))


def get_all_instance_id_names():
    instance_names = crds_manager.get(INSTANCE_NAME_CRD_SINGULAR, required=False)
    instance_name_ids = {}
    if instance_names:
        for instance_name in instance_names['items']:
            instance_name_ids[instance_name['spec']['latest-instance-id']] = instance_name['spec']['name']
    label_prefix = labels_manager.get_label_prefix()
    for instance in crds_manager.get(INSTANCE_CRD_SINGULAR)['items']:
        instance_id = instance['metadata']['labels'][f'{label_prefix}/crd-ckaninstance-name']
        instance_name = instance_name_ids.pop(instance_id, None)
        yield {'id': instance_id, 'name': instance_name}
    for instance_id, instance_name in instance_name_ids.items():
        yield {'id': instance_id, 'name': instance_name}


def get_all_instances():
    return crds_manager.get(INSTANCE_CRD_SINGULAR)['items']


def list_instances(full=False, quick=False, withCredentials=False, name=None):
    if quick:
        for instance in get_all_instance_id_names():
            if name is not None and instance['name'] != name: continue
            yield {**instance, 'ready': None}
    else:
        for instance_data in get_all_instances():
            metadata_keys = ('name',)
            spec_keys = ('id', 'siteUrl', 'siteTitle', 'domain', 'registerSubdomain')
            try:
                spec = instance_data['spec']
                for k in spec_keys:
                    instance_data[k] = spec.get(k)
                metadata = instance_data['metadata']
                for k in metadata_keys:
                    instance_data[k] = metadata.get(k)
                instance_type = instance_data['metadata']['labels'].get('{}/instance-type'.format(labels_manager.get_label_prefix()))
                deployment = deployment_manager.get(instance_data['id'], instance_type, instance_data)
                instance_data['ready'] = deployment.get('ready')
            except Exception as e:
                pass
            if not full:
                instance_data = dict(
                    (k, v)
                    for k, v in instance_data.items()
                    if k in ('ready', *spec_keys, *metadata_keys)
                )
            if withCredentials:
                instance_data['admin_password'] = config_manager.get(
                    'CKAN_ADMIN_PASSWORD',
                    secret_name='ckan-admin-password',
                    namespace=instance_data['id']
                )
            yield instance_data


def delete_name(instance_name):
    crds_manager.delete(INSTANCE_NAME_CRD_SINGULAR, instance_name)


def set_name(instance_id, instance_name, dry_run=False):
    resource = crds_manager.get(INSTANCE_NAME_CRD_SINGULAR, name=instance_name, required=False)
    if resource:
        resource['spec']['latest-instance-id'] = instance_id
        if not resource['spec']['instance-ids'].get(instance_id):
            resource['spec']['instance-ids'][instance_id] = {'added': datetime.datetime.now()}
    else:
        resource = crds_manager.get_resource(
            INSTANCE_NAME_CRD_SINGULAR, instance_name,
            spec={
                'name': instance_name,
                'latest-instance-id': instance_id,
                'instance-ids': {
                    instance_id: {'added': datetime.datetime.now()}
                }
            }
        )
    if dry_run:
        logs.print_yaml_dump(resource)
    else:
        kubectl.apply(resource)


def set_storage(instance_id, instance_name, dry_run=False):
    from ckan_cloud_operator.providers.storage.manager import get_provider, get_provider_id

    resource = crds_manager.get(INSTANCE_NAME_CRD_SINGULAR, name=instance_name, required=False)

    storage_provider = get_provider(default=None, provider_id=get_provider_id())
    bucket = storage_provider.create_bucket(instance_id, exists_ok=True, dry_run=dry_run)
    resource['spec']['ckanStorageBucket'] = {
        storage_provider.PROVIDER_ID: bucket
    }

    if dry_run:
        logs.print_yaml_dump(resource)
    else:
        kubectl.apply(resource)


def create_ckan_admin_user(instance_id_or_name, name, email=None, password=None, dry_run=False, use_paster=False):
    if not email:
        default_root_domain = routers_manager.get_default_root_domain()
        email = f'{name}@{instance_id_or_name}.{default_root_domain}'
    if not password:
        password = _generate_password(8)
    instance_id, instance_type, instance = _get_instance_id_and_type(instance_id_or_name)
    user = {
        'name': name,
        'email': email,
        'password': password
    }
    if not dry_run:
        pod_name = _get_running_pod_name(instance_id)
        name, password, email = [user[k] for k in ['name', 'password', 'email']]
        logs.info(f'Creating CKAN admin user with {name} ({email}) on pod {pod_name}')

        if use_paster:
            logs.subprocess_check_call(
                f'echo y | kubectl -n {instance_id} exec -i {pod_name} -- ckan-paster --plugin=ckan sysadmin -c /etc/ckan/production.ini add {name} password={password} email={email}',
                shell=True
            )
        else:
            logs.subprocess_check_call(
                f'echo y | kubectl -n {instance_id} exec -i {pod_name} -- ckan --config /etc/ckan/production.ini sysadmin add {name} password={password} email={email}',
                shell=True
            )

    return {
        'instance-id': instance_id,
        'instance-type': instance_type,
        **{f'admin-{k}': v for k, v in user.items()},
        **({'dry-run': True} if dry_run else {}),
    }


def delete_ckan_admin_user(instance_id_or_name, name, dry_run=False, use_paster=False):
    instance_id, instance_type, instance = _get_instance_id_and_type(instance_id_or_name)

    if not dry_run:
        pod_name = _get_running_pod_name(instance_id)
        logs.info(f'Removing CKAN admin user {name} from sys-admins')

        if use_paster:
            logs.subprocess_check_call(
                f'echo y | kubectl -n {instance_id} exec -i {pod_name} -- ckan-paster --plugin=ckan sysadmin -c /etc/ckan/production.ini remove {name}',
                shell=True
            )
        else:
            logs.subprocess_check_call(
                f'echo y | kubectl -n {instance_id} exec -i {pod_name} -- ckan --config /etc/ckan/production.ini sysadmin remove {name}',
                shell=True
            )

def run_solr_commands(instance_id_or_name, command, dataset_id='', dry_run=False, use_paster=False):
    instance_id, instance_type, instance = _get_instance_id_and_type(instance_id_or_name)

    if not dry_run:
        pod_name = _get_running_pod_name(instance_id)
        logs.info(f'Running Search Index {command}')
        if use_paster:
            answer = logs.subprocess_check_output(
                f'kubectl -n {instance_id} exec -i {pod_name} -- ckan-paster --plugin=ckan -c /etc/ckan/production.ini search-index {command}',
                shell=True
            )
            for line in str(answer).replace('\\r', '\\n').split('\\n'):
                if line:
                    logs.info(str(line))
        else:
            answer = logs.subprocess_check_output(
                f'kubectl -n {instance_id} exec -i {pod_name} -- ckan --config /etc/ckan/production.ini search-index {command} {dataset_id}',
                shell=True
            )
            for line in str(answer).replace('\\r', '\\n').split('\\n'):
                if line:
                    logs.info(str(line))


def run_ckan_commands(instance_id_or_name, command, dry_run=False, use_paster=False):
    instance_id, instance_type, instance = _get_instance_id_and_type(instance_id_or_name)
    if not dry_run:
        pod_name = _get_running_pod_name(instance_id)
        logs.info(f'Running Search Index {command}')
        if use_paster:
            answer = logs.subprocess_check_output(
                f'kubectl -n {instance_id} exec -i {pod_name} -- ckan-paster --plugin=ckan -c /etc/ckan/production.ini {command}',
                shell=True
            )
            for line in str(answer).replace('\\r', '\\n').split('\\n'):
                if line:
                    logs.info(str(line))
        else:
            answer = logs.subprocess_check_output(
                f'kubectl -n {instance_id} exec -i {pod_name} -- ckan --config /etc/ckan/production.ini {command}',
                shell=True
            )
            for line in str(answer).replace('\\r', '\\n').split('\\n'):
                if line:
                    logs.info(str(line))


def get_container_logs(instance_id, **kubectl_args):
    service = kubectl_args.pop('service')
    pod_name = _get_running_pod_name(instance_id, service=service)
    stream_logs = '-f ' if kubectl_args.pop('follow', None) else ''
    k_args = [f'--{k}={v}' for k,v in kubectl_args.items() if v is not None]
    full_args = stream_logs + ' '.join(k_args)
    _stream_logs(f'kubectl -n {instance_id} logs {pod_name} {full_args}')


def ssh_into_container(instance_id, service, command):
    pod_name = _get_running_pod_name(instance_id, service=service)
    subprocess.run(f'kubectl -n {instance_id} exec -it {pod_name} {command}', shell=True)


def _get_running_pod_name(instance_id, service='ckan'):
    pod_name = None
    while not pod_name:
        try:
            pod_name = kubectl.get_deployment_pod_name(service, instance_id, use_first_pod=True, required_phase='Running')
            break
        except Exception as e:
            logs.warning('Failed to find running ckan pod', str(e))
        time.sleep(20)
    return pod_name



def _get_instance_id_and_type(instance_id_or_name=None, instance_id=None, required=True):
    if instance_id:
        logs.debug(f'Getting instance type using instance_id', instance_id=instance_id)
        instance = crds_manager.get(INSTANCE_CRD_SINGULAR, name=instance_id, required=False)
        instance_name = None
    else:
        logs.debug(f'Attempting to get instance type using id', instance_id_or_name=instance_id_or_name)
        instance = crds_manager.get(INSTANCE_CRD_SINGULAR, name=instance_id_or_name, required=False)
        if instance:
            instance_id = instance_id_or_name
            instance_name = None
        else:
            logs.debug(f'Attempting to get instance type from instance name', instance_id_or_name=instance_id_or_name)
            instance_name = crds_manager.get(INSTANCE_NAME_CRD_SINGULAR, name=instance_id_or_name, required=False)
            if instance_name:
                instance_id = instance_name['spec'].get('latest-instance-id')
                logs.debug(instance_id=instance_id)
                instance = crds_manager.get(INSTANCE_CRD_SINGULAR, name=instance_id, required=False)
                instance_name = instance_id_or_name
            else:
                instance_name = None
    if instance:
        instance_type = instance['metadata']['labels'].get('{}/instance-type'.format(labels_manager.get_label_prefix()))
    else:
        instance_type = None
    logs.debug_yaml_dump(instance_name=instance_name, instance_id=instance_id, instance_type=instance_type,
                         instance=bool(instance))
    if required:
        assert instance_id and instance_type and len(instance) > 2, f'Failed to find instance (instance_id_or_name={instance_id_or_name}, instance_id={instance_id})'
    return instance_id, instance_type, instance


def _generate_password(length):
    return binascii.hexlify(os.urandom(length)).decode()


def _stream_logs(command):
    process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
    for c in iter(lambda: process.stdout.read(1), b''):
        sys.stdout.buffer.write(c)



import click
import binascii
import os
import yaml

from ckan_cloud_operator import logs

from . import driver


_api_endpoint, _access_key, _secret = None, None, None


def _driver_yaml_dump(method, *args, **kwargs):
    print(yaml.dump(getattr(driver, method)(_api_endpoint, _access_key, _secret, *args, **kwargs),
                    default_flow_style=False))


@click.group()
@click.option('--api-endpoint')
@click.option('--access-key')
@click.option('--secret')
def rancher(api_endpoint, access_key, secret):
    """Manage Rancher unrelated to the operator or cluster"""
    global _api_endpoint, _access_key, _secret
    _api_endpoint, _access_key, _secret = api_endpoint, access_key, secret


@rancher.command()
@click.argument('USERNAME')
@click.argument('PASSWORD')
def create_user(username, password):
    print(yaml.dump(driver.create_user(_api_endpoint, _access_key, _secret, username, password), default_flow_style=False))
    logs.exit_great_success(quiet=True)


@rancher.command()
@click.argument('CLUSTER_ID')
@click.argument('USER_ID')
@click.argument('USER_PRINCIPAL_ID')
def create_admin_role_bindings(cluster_id, user_id, user_principal_id):
    _driver_yaml_dump('create_cluster_role_template_binding', cluster_id, user_principal_id)
    print('---')
    _driver_yaml_dump('create_global_role_binding', user_id)
    logs.exit_great_success(quiet=True)


@rancher.command()
@click.argument('USERNAME')
@click.argument('PASSWORD')
def login(username, password):
    print(yaml.dump(driver.login(_api_endpoint, username, password), default_flow_style=False))
    logs.exit_great_success(quiet=True)


@rancher.command()
@click.argument('USER_LOGIN_TOKEN')
@click.argument('TOKEN_DESCRIPTION')
def create_user_token(user_login_token, token_description):
    print(yaml.dump(driver.create_user_token(_api_endpoint, user_login_token, token_description), default_flow_style=False))
    logs.exit_great_success(quiet=True)



from urllib.parse import urlparse

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.providers.routers import manager as routers_manager


def get_datapusher_url(instance_datapusher_url):
    if instance_datapusher_url and len(instance_datapusher_url) > 10:
        hostname = urlparse(instance_datapusher_url).hostname
        if hostname.endswith('.l3.ckan.io'):
            datapusher_name = hostname.replace('.l3.ckan.io', '')
        elif hostname.endswith('.ckan.io'):
            datapusher_name = hostname.replace('.ckan.io', '')
        else:
            logs.warning(f'failed to parse datapusher url from instance datapusher url: {instance_datapusher_url}')
            datapusher_name = None
        if datapusher_name:
            routes = kubectl.get(
                f'CkanCloudRoute -l ckan-cloud/route-datapusher-name={datapusher_name},ckan-cloud/route-type=datapusher-subdomain',
                required=False
            )
            if routes:
                routes = routes.get('items', [])
                if len(routes) > 0:
                    assert len(routes) == 1
                    route = routes[0]
                    sub_domain = route['spec']['sub-domain']
                    root_domain = route['spec']['root-domain']
                    assert sub_domain and sub_domain != 'default', f'invalid sub_domain: {sub_domain}'
                    if not root_domain or root_domain == 'default':
                        default_root_domain = routers_manager.get_default_root_domain()
                        assert default_root_domain, 'missing routers default root domain'
                        root_domain = default_root_domain
                    return 'https://{}.{}/'.format(sub_domain, root_domain)
                else:
                    logs.warning(f'failed to find route for datapusher: {datapusher_name}')
            else:
                logs.warning(f'failed to find route for datapusher: {datapusher_name}')
    return None



import subprocess

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs


def init(tiller_namespace_name):
    if kubectl.get('ns', tiller_namespace_name, required=False):
        logs.info('namespace already exists')
    else:
        subprocess.check_call(['kubectl', 'create', 'namespace', tiller_namespace_name])
    tiller_service_account = {
        "apiVersion": "v1",
        "kind": "ServiceAccount",
        "metadata": {
            "name": f"{tiller_namespace_name}-tiller",
            "namespace": tiller_namespace_name
        }
    }
    cluster_role_binding = {
        "apiVersion": "rbac.authorization.k8s.io/v1",
        "kind": "ClusterRoleBinding",
        "metadata": {
          "name": f"{tiller_namespace_name}-tiller"
        },
        "roleRef": {
          "apiGroup": "rbac.authorization.k8s.io",
          "kind": "ClusterRole",
          "name": "cluster-admin"
        },
        "subjects": [
            {
                "kind": "ServiceAccount",
                "name": f"{tiller_namespace_name}-tiller",
                "namespace": tiller_namespace_name
            }
        ]
    }
    kubectl.apply(tiller_service_account)
    kubectl.apply(cluster_role_binding)
    tiller_cmd = '' if _check_helm_version() == 3 else f' --tiller-namespace {tiller_namespace_name}'
    subprocess.check_call(
        f'helm init --upgrade --stable-repo-url https://charts.helm.sh/stable --service-account {tiller_namespace_name}-tiller --history-max 10' + tiller_cmd,
        shell=True
    )


def deploy(tiller_namespace, chart_repo, chart_name, chart_version, release_name, values_filename=None, namespace=None,
           dry_run=False, chart_repo_name=None, values=None, service_account=None):
    if not chart_repo_name:
        chart_repo_name = 'ckan-cloud'
        logs.info(chart_repo_name=chart_repo_name)
    if chart_repo:
        subprocess.check_call(f'helm repo add "{chart_repo_name}" "{chart_repo}"', shell=True)
    assert not (values_filename and values), 'Only one of `values_filename` or `values` should be passed'

    version_args = f'--version "{chart_version}"' if chart_version else ''
    dry_run_args = '--dry-run --debug'
    tiller_cmd = '' if _check_helm_version() == 3 else f' --tiller-namespace {tiller_namespace}'
    cmd = f'helm upgrade {release_name} {chart_name} ' \
          f' --install --namespace "{namespace}" -i {version_args}'
    if values_filename:
        cmd += f' -f {values_filename}'
    elif values:
        for key, value in values.items():
            cmd += f' --set {key}={value}'
    cmd += tiller_cmd
    logs.info('Running helm upgrade --dry-run')
    if logs.CKAN_CLOUD_OPERATOR_DEBUG_FILE:
        logs.info(f'helm dry run debug output is written to debug log file: {logs.CKAN_CLOUD_OPERATOR_DEBUG_FILE}')
        with open(logs.CKAN_CLOUD_OPERATOR_DEBUG_FILE, 'a') as f:
            subprocess.check_call(f'{cmd} {dry_run_args}', shell=True, stdout=f, stderr=subprocess.STDOUT)
    else:
        subprocess.check_call(f'{cmd} {dry_run_args}', shell=True)
    if not dry_run:
        logs.info('Running helm upgrade for real this time')
        subprocess.check_call(cmd, shell=True)


def delete(tiller_namespace, release_name):
    tiller_cmd = '' if _check_helm_version() == 3 else f' --tiller-namespace {tiller_namespace}'
    subprocess.check_call(f'helm delete --purge --timeout 5 {release_name}' + tiller_cmd, shell=True)


def check_status(instance_id):
    subprocess.check_call(f'helm status ckan-cloud-{instance_id} -n {instance_id}', shell=True)

def get_values(instance_id):
    return subprocess.check_output(f'helm get values ckan-cloud-{instance_id} -n {instance_id} -o json', shell=True)


def _check_helm_version():
    return 3 if 'v3.' in str(subprocess.check_output('helm version -c', shell=True)) else 2



import click

from .helm import cli as helm_cli
from .drone import cli as drone_cli
from .drone import manager
from ckan_cloud_operator.providers.ckan.deployment import manager as deployment_manager

drone_manager = manager.Drone()

@click.group()
def deployment():
    """Manage CKAN instance deployments"""
    pass

@click.group()
def image():
    """Manage CKAN instance deployments"""
    pass

@deployment.command()
@click.option('--branch', default='develop', help='Source Branch for build [default: develop]')
def logs(branch):
    """See CKAN instances deployment Logs"""
    drone_manager.initialize()
    drone_manager.builds_logs(branch)


@deployment.command()
@click.argument('instance-id')
def status(instance_id):
    """See CKAN instances deployment status"""
    helm_driver.check_status(instance_id)


@deployment.command()
@click.argument('instance-id')
def version(instance_id):
    """See CKAN instances deployment version"""
    deployment_manager.get_deployment_version(instance_id)


@image.command()
@click.argument('instance-id')
@click.option('--service', default='ckan', help='Source Branch for build [default: develop]')
def get(instance_id, service):
    """Get instances container image"""
    deployment_manager.get_image(instance_id, service=service)


@image.command()
@click.argument('instance-id')
@click.argument('image-name')
@click.option('--service', default='ckan', help='Source Branch for build [default: develop]')
def set(instance_id, image_name, service):
    """Set instances container image"""
    deployment_manager.set_image(instance_id, image_name, service=service)


deployment.add_command(helm_cli.helm)
deployment.add_command(drone_cli.drone)
deployment.add_command(image)



from ckan_cloud_operator import kubectl


def update_service_account(service_account_name, labels, namespace=None):
    service_account = kubectl.get_resource(
        'v1', 'ServiceAccount', service_account_name, labels
    )
    if namespace:
        service_account['metadata']['namespace'] = namespace
    kubectl.apply(service_account)


def update_cluster_role(name, rules, labels):
    kubectl.apply({
        "apiVersion": "rbac.authorization.k8s.io/v1",
        "kind": "ClusterRole",
        "metadata": {
            "name": name,
            'labels': labels
        },
        "rules": rules
    }, reconcile=True)


def update_admin_cluster_role(name, labels):
    update_cluster_role(name, [
        {
            "apiGroups": [
                "*"
            ],
            "resources": [
                "*"
            ],
            "verbs": [
                "*"
            ]
        },
        {
            "nonResourceURLs": [
                "*"
            ],
            "verbs": [
                "*"
            ]
        }
    ], labels)


def update_cluster_role_binding(name, subject,
                                cluster_role_name, labels):
    resource = {
        'apiVersion': 'rbac.authorization.k8s.io/v1',
        'kind': 'ClusterRoleBinding',
        'metadata': {
            'name': name,
            'labels': labels
        },
        'subjects': [subject],
        'roleRef': {
            'kind': 'ClusterRole',
            'name': cluster_role_name,
            'apiGroup': 'rbac.authorization.k8s.io'
        }
    }
    kubectl.apply(resource, reconcile=True)


def update_role(name, labels, rules, namespace):
    kubectl.apply({
        "apiVersion": "rbac.authorization.k8s.io/v1",
        "kind": "Role",
        "metadata": {
            "name": name,
            'namespace': namespace,
            'labels': labels
        },
        "rules": rules
    }, reconcile=True)


def update_role_binding(name, role_name, namespace, service_account_name, labels):
    kubectl.apply({
        'apiVersion': 'rbac.authorization.k8s.io/v1',
        'kind': 'RoleBinding',
        'metadata': {
            'name': name,
            'namespace': namespace,
            'labels': labels
        },
        'subjects': [
            {
                'kind': 'ServiceAccount',
                'name': service_account_name,
                'namespace': namespace,
            }
        ],
        'roleRef': {
            'kind': 'Role',
            'name': role_name,
            'namespace': namespace,
            'apiGroup': 'rbac.authorization.k8s.io'
        }
    }, reconcile=True)


def get_kubeconfig(cluster_name, service_account_namespace, service_account_name, cluster_spec=None):
    service_account = kubectl.get(f'ServiceAccount {service_account_name}', namespace=service_account_namespace)
    secret_name = service_account['secrets'][0]['name']
    secret = kubectl.decode_secret(kubectl.get(f'secret {secret_name}', namespace=service_account_namespace))
    config = kubectl.get('config view', get_cmd='')
    assert len(config['clusters']) == 1
    if not cluster_spec:
        cluster_spec = {"server": config['clusters'][0]['cluster']['server']}
    return {
        "apiVersion": "v1",
        "kind": "Config",
        "users": [
            {
                "name": service_account_name,
                "user": {
                    "token": secret['token']
                }
            }
        ],
        "clusters": [
            {
                "name": cluster_name,
                "cluster": cluster_spec
            }
        ],
        "contexts": [
            {
                "name": cluster_name,
                "context": {
                    "cluster": cluster_name,
                    "user": service_account_name
                }
            }
        ],
        "current-context": cluster_name
    }


from ckan_cloud_operator import kubectl


DEFAULT_CHART_VALUES = {
    'chart-repo': 'https://raw.githubusercontent.com/ViderumGlobal/ckan-cloud-helm/master/charts_repository',
    'chart-repo-name': 'ckan-cloud',
    'chart-name': 'ckan-cloud/elk',
    'chart-version': 'v0.0.2',
}

DEFAULT_VALUES = {
    'elasticsearchMemoryLimit': '3Gi',
    'elasticsearchStorage': '300Gi',
    'elsaticsearchReplicas': 2,
    'logstashMemoryLimit': '1.5Gi',
    'storageclassProvisioner': 'kubernetes.io/aws-ebs'
}


def pre_update_hook(instance_id, instance, res, sub_domain, root_domain, modify_spec_callback):
    modify_spec_callback(lambda i: i.update(**{
        k: v for k, v in DEFAULT_CHART_VALUES.items()
        if not instance['spec'].get(k)
    }))
    modify_spec_callback(lambda i: i.setdefault('values', {}).update(**{
        k: v for k, v in DEFAULT_VALUES.items()
        if not instance['spec'].get('values', {}).get(k)
    }))


def pre_deploy_hook(instance_id, instance, deploy_kwargs):
    pass


def post_deploy_hook(instance_id, instance, deploy_kwargs):
    pass


def pre_delete_hook(instance_id, instance, delete_kwargs):
    pass


def post_delete_hook(instance_id, instance, delete_kwargs):
    pass


def get_backend_url(instance_id, instance, backend_url):
    return 'http://kibana.elk:80'


def get(instance_id, instance, res):
    res['ready'] = True
    app_pods_status = {}
    for pod in kubectl.get('pods', namespace=instance_id, required=True)['items']:
        app = pod['metadata']['labels'].get('app')
        if not app:
            app = 'unknown'
        item_status = kubectl.get_item_detailed_status(pod)
        if item_status.get('errors') and len(item_status['errors']) > 0:
            res['ready'] = False
        app_pods_status.setdefault(app, {})[pod['metadata']['name']] = item_status
    app_deployments_status = {}
    for deployment in kubectl.get('deployments', namespace=instance_id, required=True)['items']:
        app = deployment['metadata']['labels'].get('app')
        if not app:
            app = 'unknown'
        item_status = kubectl.get_item_detailed_status(deployment)
        if item_status.get('errors') and len(item_status['errors']) > 0:
            res['ready'] = False
        app_deployments_status.setdefault(app, {})[deployment['metadata']['name']] = item_status
    if 'api' not in app_pods_status or 'ui' not in app_pods_status:
        res['ready'] = False
    res['app'] = {
        'pods': app_pods_status,
        'deployments': app_deployments_status
    }



import click

from . import manager


@click.group()
def web_ui():
    """Manage the centralized db web-ui"""
    pass


@web_ui.command()
def initialize():
    manager.initialize()


@web_ui.command()
def start():
    manager.start()



import requests

from ckan_cloud_operator import logs


def post_json(api_endpoint, access_key, secret, path, data):
    if access_key or secret:
        assert api_endpoint.startswith('https://'), f'invalid endpoint: {api_endpoint}'
        endpoint = api_endpoint.replace('https://', '')
        url = f'https://{access_key}:{secret}@{endpoint}{path}'
    else:
        url = f'{api_endpoint}{path}'
    res = requests.post(url, json=data)
    assert res.status_code in [200, 201], f'invalid status_code: {res.status_code}\n{res.text}'
    return res.json()


def create_user(api_endpoint, access_key, secret, username, password):
    return post_json(
        api_endpoint, access_key, secret,
        '/user',
        {
            "enabled": True, "me": False, "mustchangePassword": False,
            "password": password, "type": "user", "username": username
        }
    )


def create_cluster_role_template_binding(api_endpoint, access_key, secret, cluster_id, user_principal_id):
    return post_json(
        api_endpoint, access_key, secret,
        '/clusterroletemplatebinding',
        {
            "type": "clusterRoleTemplateBinding",
            "clusterId":cluster_id,
            "userPrincipalId": user_principal_id,
            "roleTemplateId": "cluster-owner"
        }
    )


def create_global_role_binding(api_endpoint, access_key, secret, user_id):
    return post_json(
        api_endpoint, access_key, secret,
        '/globalrolebinding',
        {
            "type": "globalRoleBinding",
            "globalRoleId": "admin",
            "userId": user_id,
            "subjectKind": "User"
        }
    )


def login(api_endpoint, username, password):
    return post_json(
        api_endpoint, None, None,
        '-public/localProviders/local?action=login', {
            "username": username,
            "password": password
        }
    )


def create_user_token(api_endpoint, user_login_token, token_description):
    access_key, secret = user_login_token.split(':')
    return post_json(
        api_endpoint, access_key, secret,
        '/token',
        {
            "current": False,
            "description": token_description,
            "expired": False,
            "isDerived": False,
            "ttl": 0,
            "type": "token"
        }
    )



import click
import time
import yaml

from ckan_cloud_operator import logs

from . import manager as db_manager

from .gcloudsql import cli as gcloudsql_cli
from .proxy import cli as proxy_cli
from .web_ui import cli as web_ui_cli


@click.group('db')
def db_group():
    """Manage the centralized db"""
    pass


db_group.add_command(gcloudsql_cli.gcloudsql_group, 'gcloudsql')
db_group.add_command(proxy_cli.proxy)
db_group.add_command(web_ui_cli.web_ui)


@db_group.command()
@click.option('--interactive', is_flag=True)
def initialize(interactive):
    db_manager.initialize(interactive=interactive)


@db_group.command()
@click.option('--admin', is_flag=True)
@click.option('--deis-instance')
@click.option('--datastore', is_flag=True)
@click.option('--datastore-ro', is_flag=True)
@click.option('--db-prefix')
def connection_string(admin, deis_instance, datastore, datastore_ro, db_prefix):
    """Get a DB connection string

    Example: psql -d $(ckan-cloud-operator db connection-string --admin)
    """
    if deis_instance:
        print(db_manager.get_deis_instsance_external_connection_string(deis_instance, is_datastore=datastore,
                                                                       is_datastore_readonly=datastore_ro,
                                                                       admin=admin, db_prefix=db_prefix))
    elif admin:
        assert not deis_instance and not datastore and not datastore_ro
        print(db_manager.get_external_admin_connection_string(db_prefix=db_prefix))
    else:
        raise Exception('invalid arguments')


@db_group.command()
def get_all_dbs_users():
    dbs, users = db_manager.get_all_dbs_users()
    print(yaml.dump({
        'dbs': [' | '.join(map(str, db)) for db in dbs],
        'users': [' | '.join(map(str, user)) for user in users]
    }, default_flow_style=False))



import click
import yaml

from . import driver
from ckan_cloud_operator import logs


@click.group()
def jenkins():
    """Interact with a Jenkins server"""
    pass


@jenkins.command()
@click.argument('JENKINS_USER')
@click.argument('JENKINS_TOKEN')
@click.argument('JENKINS_URL')
@click.argument('POST_JSON_DATA', required=False)
def curl(jenkins_user, jenkins_token, jenkins_url, post_json_data):
    logs.print_yaml_dump(driver.curl(jenkins_user, jenkins_token, jenkins_url, post_json_data))
    logs.exit_great_success(quiet=True)



import click
import time

from ckan_cloud_operator import logs

from ckan_cloud_operator.providers.cluster import manager
from ckan_cloud_operator.labels import manager as labels_manager

from .aws.cli import aws as aws_cli


@click.group()
def cluster():
    """Manage the cluster"""
    pass


@cluster.command()
@click.option('--interactive', is_flag=True)
@click.option('--cluster-provider', default='gcloud')
@click.option('--skip-to')
def initialize(interactive, cluster_provider, skip_to):
    """Initialize the currently connected cluster"""
    manager.initialize(interactive=interactive, default_cluster_provider=cluster_provider, skip_to=skip_to)
    logs.exit_great_success()



############################################################################
## Commands Below are not used for now and are making unnecessary noise   ##
## I'm commenting them for now, but not deleting as they might be useful  ##
## In future. Will bring them back as needed                              ##
############################################################################


# cluster.add_command(aws_cli)
#
# @cluster.command()
# @click.option('--debug', is_flag=True)
# @click.option('--full', is_flag=True)
# def info(debug, full):
#     manager.print_info(debug=debug, minimal=not full)
#
#
# @cluster.command()
# @click.argument('DISK_SIZE_GB')
# @click.argument('ZONE', required=False, default=0)
# def create_volume(disk_size_gb, zone):
#     label_prefix = labels_manager.get_label_prefix()
#     print(manager.create_volume(
#         disk_size_gb,
#         {f'{label_prefix}/operator-volume-source': 'cli'},
#         zone=zone
#     ))
#
#
# @cluster.command()
# @click.argument('COMMAND', nargs=-1)
# def provider_exec(command):
#     manager.provider_exec(' '.join(command))
#
#
# @cluster.command()
# @click.option('--expander', default='random', help='random|most-pods|least-waste|price (default: random)')
# @click.option('--min-nodes', type=int, help='specifies the minimum number of nodes for the node pool [gcloud provider only]')
# @click.option('--max-nodes', type=int, help='specifies the maximum number of nodes for the node pool [gcloud provider only]')
# @click.option('--zone', help='specifies the cluster\'s compute zone [gcloud provider only]')
# @click.option('--node-pool', help='specifies the desired node pool. If you have only one node pool, supply default-pool to this flag. [gcloud provider only]')
# def setup_autoscaler(expander, min_nodes, max_nodes, zone, node_pool):
#     """Set up the cluster autoscaler from stable/cluster-autoscaler helm package"""
#     manager.setup_autoscaler(expander, min_nodes, max_nodes, zone, node_pool)



PROVIDER_SUBMODULE='ckan'

INSTANCE_CRD_SINGULAR = 'ckaninstance'
INSTANCE_CRD_PLURAL_SUFFIX = 'ckaninstances'
INSTANCE_CRD_KIND_SUFFIX = 'CkanInstance'

INSTANCE_NAME_CRD_SINGULAR = 'ckaninstancename'
INSTANCE_NAME_CRD_PLURAL_SUFFIX = 'ckaninstancenames'
INSTANCE_NAME_CRD_KIND_SUFFIX = 'CkanInstanceName'



from ckan_cloud_operator import logs
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.routers import manager as routers_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager


def initialize(interactive=False):
    config_manager.interactive_set(
        {
            'env-id': None,
            'default-root-domain': None,
            'dns-provider': 'route53' if cluster_manager.get_provider_id() == 'aws' else None
        },
        configmap_name='routers-config',
        interactive=interactive
    )
    dns_provider = get_dns_provider()
    if dns_provider.lower() == 'none':
        if cluster_manager.get_provider_id() != 'minikube':
            return
    logs.info(dns_provider=dns_provider)
    if dns_provider == 'cloudflare':
        config_manager.interactive_set(
            {
                'cloudflare-email': None,
                'cloudflare-api-key': None
            },
            configmap_name='routers-config',
            interactive=interactive
        )
    routers_manager.install_crds()
    infra_router_name = routers_manager.get_default_infra_router_name()
    default_root_domain = config_manager.get('default-root-domain', configmap_name='routers-config', required=True)
    logs.info('Creating infra router', infra_router_name=infra_router_name, default_root_domain=default_root_domain)
    routers_manager.create(
        infra_router_name,
        routers_manager.get_traefik_router_spec(
            default_root_domain,
            config_manager.get('cloudflare-email', configmap_name='routers-config', required=False, default=None),
            config_manager.get('cloudflare-api-key', configmap_name='routers-config', required=False, default=None),
            dns_provider=dns_provider
        )
    )


def get_env_id():
    return config_get('env-id') or 'p'


def get_dns_provider():
    return config_manager.get(key='dns-provider', configmap_name='routers-config')


def get_default_root_domain():
    return config_get('default-root-domain')


def config_get(key):
    return config_manager.get(key, configmap_name='routers-config')


def get_cloudflare_credentials():
    return (
        config_manager.get('cloudflare-email', configmap_name='routers-config'),
        config_manager.get('cloudflare-api-key', configmap_name='routers-config')
    )


def update_dns_record(dns_provider, sub_domain, root_domain, load_balancer_ip_or_hostname, cloudflare_email=None,
                      cloudflare_auth_key=None):
    logs.info('updating DNS record', dns_provider=dns_provider, sub_domain=sub_domain, root_domain=root_domain,
              load_balancer_ip_or_hostname=load_balancer_ip_or_hostname,
              cloudflare_email=cloudflare_email,
              cloudflare_auth_key_len=len(cloudflare_auth_key) if cloudflare_auth_key else 0)
    if dns_provider == 'cloudflare':
        from ckan_cloud_operator import cloudflare
        cloudflare.update_a_record(cloudflare_email, cloudflare_auth_key, root_domain,
                                   f'{sub_domain}.{root_domain}', load_balancer_ip_or_hostname)
    elif dns_provider == 'route53':
        from ckan_cloud_operator.providers.cluster.aws import manager as aws_manager
        aws_manager.update_dns_record(sub_domain, root_domain, load_balancer_ip_or_hostname)
    elif dns_provider == 'azure':
        from ckan_cloud_operator.providers.cluster.azure import manager as azure_manager
        azure_manager.create_dns_record(sub_domain, root_domain, load_balancer_ip_or_hostname)
    elif dns_provider.lower() == 'none':
        return
    else:
        raise NotImplementedError()
